--|
--| Filename        : $Source: /home/erdmann/MySQL/RCS/db.adb,v $
--| Description     : Database interface to MySQL
--| Author          : Michael Erdmann
--| Created On      : 25.3.1999
--| Last Modified By: $Author: erdmann $
--| Last Modified On: $Date: 2000/11/27 21:08:14 $
--| Status          : $State: Exp $
--|
--| Copyright (C) 2000 Michael Erdmann
--|
--| This program is free software; you can redistribute it and/or
--| modify it under the terms of the GNU General Public License
--| as published by the Free Software Foundation; either version 2
--| of the License, or (at your option) any later version.
--|
--| This program/code is distributed in the hope that it will be useful,
--| but WITHOUT ANY WARRANTY; without even the implied warranty of
--| MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
--| GNU General Public License for more details.
--|
--| You should have received a copy of the GNU General Public License
--| along with this program; if not, write to the Free Software
--| Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
--|
--| Functional Description
--| ======================
--| This components implements a thick binding to the MySQL product
--| which is available for the Linux Plattform.
--|
--|
--| Restrictions
--| ============
--| Only 100 fields per query
--| Number of Queries per connection is 10.
--|
--| References
--| ==========
--|
--|
--| History
--| =======
--| $Log: db.adb,v $
--| Revision 1.1  2000/11/27 21:08:14  erdmann
--| resource of ASCL code done
--|
--| Revision 1.3  2000/05/14 19:33:07  acde
--| Put under GPL
--|
--| Revision 1.2  2000/04/23 10:48:14  acde
--| No comments
--|
--|

--* Ada
with System;                  use System;
with System.Storage_Elements; use System.Storage_Elements;
with System.Address_To_Access_Conversions;

with Ada.Exceptions;          use Ada.Exceptions;
with Ada.Strings.Unbounded;   use Ada.Strings.Unbounded;
with Ada.Unchecked_Conversion;
with Ada.Text_IO;             use Ada.Text_IO;
with Ada.Characters.Latin_1;  use Ada.Characters;
with Ada.Calendar;            use Ada.Calendar;
with Ada.Strings.Fixed;       use Ada.Strings.Fixed;

with Interfaces.C;            use Interfaces.C;
use  Interfaces;
with Interfaces.C.Strings;    use Interfaces.C.Strings;

with Unchecked_Deallocation;

package body DB is

   Version : constant String := "$Id: db.adb,v 1.1 2000/11/27 21:08:14 erdmann Exp erdmann $";

   --|
   --| Interface to the MYSQL Library. Some of the data types are simply
   --| Addresses because the internals of these strictured do not have
   --| to be known here.
   --|
   subtype MYSQL_SERVER is  System.Address;
   subtype MYSQL_ROW    is  System.Address;
   subtype MYSQL_RES    is  System.Address;
   subtype MYSQL_FNO    is  C.unsigned;

   type MYSQL_RES_Record is                                     -- mysql.h:89
        record
            row_count    : c.unsigned;                          -- mysql.h:90
            field_count  : c.unsigned;                          -- mysql.h:91
            current_field: c.unsigned;                          -- mysql.h:91
            fields       : Address;                             -- mysql.h:92
            data         : Address;                             -- mysql.h:93
            data_cursor  : Address;                             -- mysql.h:94
            field_alloc  : Address;                             -- mysql.h:95
            row          : MYSQL_ROW;                           -- mysql.h:96
            current_row  : MYSQL_ROW;                           -- mysql.h:97
            lengths      : Address;                             -- mysql.h:98
            handle       : Address;                             -- mysql.h:99
            eof          : C.char;                              -- mysql.h:100
        end record;

   pragma Convention(C, MYSQL_RES_Record );


   function mysql_connect( mysql : MYSQL_SERVER;
                           host  : Strings.chars_ptr;
                           user  : Strings.chars_ptr;
                           passwd: Strings.chars_ptr)
                                   return MYSQL_SERVER;
   pragma Import(C, mysql_connect, "mysql_connect");

   procedure mysql_close(sock: MYSQL_SERVER);
   pragma Import(C, mysql_close, "mysql_close");

   function mysql_select_db(mysql: MYSQL_SERVER; db : Strings.chars_ptr) return c.int;
   pragma Import(C, mysql_select_db, "mysql_select_db");

   function mysql_query(mysql: MYSQL_SERVER; q : Strings.chars_ptr) return c.int;
   pragma Import(C, mysql_query, "mysql_query");

   function mysql_fetch_row(mysql: MYSQL_RES) return MYSQL_ROW;
   pragma Import(C, mysql_fetch_row, "mysql_fetch_row");

   function mysql_field_seek(mysql : MYSQL_RES; offset: MYSQL_FNO) return MYSQL_FNO;
   pragma Import(C, mysql_field_seek, "mysql_field_seek");    -- mysql.h:166

   type MYSQL_field_types is (                                  -- mysql_com.h:64
        FIELD_TYPE_DECIMAL,                                     -- mysql_com.h:64
        FIELD_TYPE_TINY,                                        -- mysql_com.h:64
        FIELD_TYPE_SHORT,                                       -- mysql_com.h:65
        FIELD_TYPE_LONG,                                        -- mysql_com.h:65
        FIELD_TYPE_FLOAT,                                       -- mysql_com.h:66
        FIELD_TYPE_DOUBLE,                                      -- mysql_com.h:66
        FIELD_TYPE_NULL,                                        -- mysql_com.h:67
        FIELD_TYPE_TIMESTAMP,                                   -- mysql_com.h:67
        FIELD_TYPE_LONGLONG,                                    -- mysql_com.h:68
        FIELD_TYPE_INT24,                                       -- mysql_com.h:68
        FIELD_TYPE_DATE,                                        -- mysql_com.h:69
        FIELD_TYPE_TIME,                                        -- mysql_com.h:69
        FIELD_TYPE_DATETIME,                                    -- mysql_com.h:70
        FIELD_TYPE_YEAR,                                        -- mysql_com.h:70
        FIELD_TYPE_NEWDATE,                                     -- mysql_com.h:71
        FIELD_TYPE_ENUM,                                        -- mysql_com.h:72
        FIELD_TYPE_SET,                                         -- mysql_com.h:73
        FIELD_TYPE_TINY_BLOB,                                   -- mysql_com.h:74
        FIELD_TYPE_MEDIUM_BLOB,                                 -- mysql_com.h:75
        FIELD_TYPE_LONG_BLOB,                                   -- mysql_com.h:76
        FIELD_TYPE_BLOB,                                        -- mysql_com.h:77
        FIELD_TYPE_VAR_STRING,                                  -- mysql_com.h:78
        FIELD_TYPE_STRING                                       -- mysql_com.h:80
    );
    for MYSQL_field_types use (                             -- mysql_com.h:64
        FIELD_TYPE_DECIMAL => 0,                                -- mysql_com.h:64
        FIELD_TYPE_TINY => 1,                                   -- mysql_com.h:64
        FIELD_TYPE_SHORT => 2,                                  -- mysql_com.h:65
        FIELD_TYPE_LONG => 3,                                   -- mysql_com.h:65
        FIELD_TYPE_FLOAT => 4,                                  -- mysql_com.h:66
        FIELD_TYPE_DOUBLE => 5,                                 -- mysql_com.h:66
        FIELD_TYPE_NULL => 6,                                   -- mysql_com.h:67
        FIELD_TYPE_TIMESTAMP => 7,                              -- mysql_com.h:67
        FIELD_TYPE_LONGLONG => 8,                               -- mysql_com.h:68
        FIELD_TYPE_INT24 => 9,                                  -- mysql_com.h:68
        FIELD_TYPE_DATE => 10,                                  -- mysql_com.h:69
        FIELD_TYPE_TIME => 11,                                  -- mysql_com.h:69
        FIELD_TYPE_DATETIME => 12,                              -- mysql_com.h:70
        FIELD_TYPE_YEAR => 13,                                  -- mysql_com.h:70
        FIELD_TYPE_NEWDATE => 14,                               -- mysql_com.h:71
        FIELD_TYPE_ENUM => 247,                                 -- mysql_com.h:72
        FIELD_TYPE_SET => 248,                                  -- mysql_com.h:73
        FIELD_TYPE_TINY_BLOB => 249,                            -- mysql_com.h:74
        FIELD_TYPE_MEDIUM_BLOB => 250,                          -- mysql_com.h:75
        FIELD_TYPE_LONG_BLOB => 251,                            -- mysql_com.h:76
        FIELD_TYPE_BLOB => 252,                                 -- mysql_com.h:77
        FIELD_TYPE_VAR_STRING => 253,                           -- mysql_com.h:78
        FIELD_TYPE_STRING => 254                                -- mysql_com.h:80
    );
   for MYSQL_field_types'size use 32;

   type st_mysql_field is                                       -- mysql.h:39
        record
            name      : Strings.chars_ptr;                      -- mysql.h:40
            table     : Strings.chars_ptr;                      -- mysql.h:41
            def       : Strings.chars_ptr;                      -- mysql.h:42
            c_type    : MYSQL_field_types;                      -- mysql.h:43
            length    : c.unsigned;                             -- mysql.h:44
            max_length: c.unsigned;                             -- mysql.h:45
            flags     : c.unsigned;                             -- mysql.h:46
            decimals  : c.unsigned;                             -- mysql.h:47
        end record;

   pragma Convention(C, st_mysql_field);                        -- mysql.h:39
   type MYSQL_Field is access all st_mysql_field;

    type enum_mysql_status is (                                 -- mysql.h:67
        MYSQL_STATUS_READY,                                     -- mysql.h:67
        MYSQL_STATUS_GET_RESULT,                                -- mysql.h:67
        MYSQL_STATUS_USE_RESULT                                 -- mysql.h:68
    );
    for enum_mysql_status'size use 32;                          -- mysql.h:67

    type struct_st_net is record
            fd          : c.int;                               -- mysql_com.h:54
            fcntl       : c.int;                               -- mysql_com.h:55
            buff        : Strings.chars_ptr;                             -- mysql_com.h:56
            buff_end    : Strings.chars_ptr;                             -- mysql_com.h:56
            write_pos   : Strings.chars_ptr;                             -- mysql_com.h:56
            last_error  : C.char_Array(0..199);             -- mysql_com.h:57
            last_errno  : c.unsigned;                       -- mysql_com.h:58
            max_packet  : c.unsigned;                       -- mysql_com.h:58
            timeout     : c.unsigned;                       -- mysql_com.h:58
            pkt_nr      : c.unsigned;                       -- mysql_com.h:58
            error       : c.signed_char;                    -- mysql_com.h:59
            return_errno: c.signed_char;                              -- mysql_com.h:59
        end record;

    pragma Convention(C,  struct_st_net);                       -- mysql_com.h:53
    subtype NET is struct_st_net;

    type A_MYSQL_FIELD_T is access all st_mysql_field;         -- mysql.h:82

    type st_mem_root is                                  -- mysql.h:18
        record
            free         : Address;                            -- mysql.h:19
            used         : Address;                             -- mysql.h:20
            min_malloc   : c.unsigned;                          -- mysql.h:21
            block_size   : c.unsigned;                          -- mysql.h:22
            error_handler: Address     ;                        -- mysql.h:23
        end record;

    pragma Convention(C, st_mem_root);                   -- mysql.h:18
    subtype MEM_ROOT is st_mem_root;                     -- mysql.h:24


    type st_mysql is  record
            The_net            : NET;                           -- mysql.h:71
            host               : Strings.chars_ptr;                       -- mysql.h:72
            user               : Strings.chars_ptr;                       -- mysql.h:72
            passwd             : Strings.chars_ptr;                       -- mysql.h:72
            unix_socket        : Strings.chars_ptr;                       -- mysql.h:72
            server_version     : Strings.chars_ptr;                       -- mysql.h:72
            host_info          : Strings.chars_ptr;                       -- mysql.h:72
            info               : Strings.chars_ptr;                       -- mysql.h:73
            db                 : Strings.chars_ptr;                       -- mysql.h:73
            port               : c.unsigned;                -- mysql.h:74
            client_flag        : c.unsigned;                -- mysql.h:74
            server_capabilities: c.unsigned;                -- mysql.h:74
            protocol_version   : c.unsigned;                -- mysql.h:75
            field_count        : c.unsigned;                -- mysql.h:76
            thread_id          : c.unsigned;                -- mysql.h:77
            affected_rows      : c.unsigned;                -- mysql.h:78
            insert_id          : c.unsigned;                -- mysql.h:79
            extra_info         : c.unsigned;                -- mysql.h:80
            status             : enum_mysql_status;             -- mysql.h:81
            fields             : A_MYSQL_FIELD_T;               -- mysql.h:82
            field_alloc        : MEM_ROOT;                      -- mysql.h:83
            free_me            : c.signed_char;                 -- mysql.h:84
            reconnect          : c.signed_char;                       -- mysql.h:85
        end record;

               -- mysql.h:89 
   function mysql_init( 
      mysql :  MYSQL_SERVER ) return MYSQL_SERVER;          
   pragma Import(C, mysql_init, "mysql_init");   

               -- mysql.h:89

   function mysql_fetch_field(
      handle: MYSQL_RES) return MYSQL_FIELD;
   pragma Import(C, mysql_fetch_field, "mysql_fetch_field");

   function mysql_store_result(
      mysql: MYSQL_SERVER) return MYSQL_RES; -- mysql.h:161
   pragma Import(C, mysql_store_result, "mysql_store_result"); -- mysql.h:161

   function mysql_use_result(
      mysql: MYSQL_SERVER) return MYSQL_RES;                   -- mysql.h:162
   pragma Import(C, mysql_use_result, "mysql_use_result");     -- mysql.h:162

   procedure mysql_free_result(
      result: MYSQL_RES);             			       -- mysql.h:163
   pragma Import(C, mysql_free_result, "mysql_free_result");   

   ---====================================================================---
   ---===                O B J  E C T     D A T A                      ===---
   ---====================================================================---

   type Field_Info_Array is array( Field_Number ) of St_MySql_Field;

   --|
   --| This is the instance of the component internal data.
   --|
   type Query_Context is record
         In_Use        : Boolean   := False;
         Result        : MYSQL_RES := Null_Address;
         Row           : MYSQL_ROW := Null_Address;
         Field         : Field_Info_Array;
         Affected_Rows : Integer := 0;
      end record;

   type Query_Table is array( Query_ID ) of Query_Context;

   type Object_Data is record
         Server_Name : Unbounded_String := Null_Unbounded_String;
         Login_Name  : Unbounded_String := Null_Unbounded_String;
         Password    : Unbounded_String := Null_Unbounded_String;

         Connected   : Boolean          := False;
         Server      : MYSQL_Server     := Null_Address;
         Query       : Query_Table;
      end record;

   ---=====================================================================---
   ---===         L O C A L   S U P P O R T   P R O C E D U R E S       ===---
   ---=====================================================================---

   --|
   --| Get pointer to object data
   --|
   function Instance(
      this : Object; info : String ) return  Object_Data_Access is
   begin
      if this.Data = null then
         raise Not_Initialized;
      end if;

      return this.Data;
   end Instance;

   --|
   --| Retrieve the address from an array fo addresses
   --|
   function Get_Address(
      row : MYSQL_ROW;
      pos : Field_Number ) return Address is
      ---
      offset : Storage_Offset :=
                   Storage_Offset( (Pos-1) * Address'Size/Storage_Element'Size );

      package Ptr_To_Address is
                new System.Address_To_Access_Conversions( Object => Address );
      use Ptr_To_Address;
      ----
   begin
      return To_Pointer( row + offset ).all ;
   end Get_Address;

   --|
   --| Copy a string from the memory address. This code is not nice but
   --| is works.
   --|
   function Copy_String (
      addr   : Address;
      len    : Positive ) return String is
      ---
      result : String( 1..len ) := ( others => ' ' );
      byte   : Character;

      package Byte_String is
                    new System.Address_To_Access_Conversions ( Object => Character );
      use Byte_String;
      ---
   begin
      for i in 1..len loop
         byte := To_Pointer( addr + Storage_Offset( i-1 ) ).all;
         if byte /= Latin_1.NUL then
            result(i) := byte;
         else
            exit;
         end if;
      end loop;

      -- Put_Line("Copy string :" & result ); 
      return result;
   end Copy_String;

   --|
   --| Verify if data and query are valid informations.
   --|
   procedure Validate(
      data   : in Object_Data_Access;
      id     : in Query_ID ) is
   begin
      if not Data.Query(id).In_Use then
         raise Usage_Error;
      end if;
   end Validate;

   --|
   --|  Return the number of rows found during a selection
   --|
   function NumRows(
      result : Address  ) return Integer is
      ---
      package RES_Access is
         new System.Address_To_Access_Conversions( Object=>MYSQL_RES_Record );
      use RES_Access;

      ptr : Object_Pointer := null;
      ---
   begin
      if result = Null_Address then
         return 0;
      end if;

      ptr := To_Pointer( result );
      return Integer(ptr.Row_Count);
   end NumRows;


   function Num_Affected(
      server : Address  ) return Integer is
      ---
      package MYSQL_Access is
         new System.Address_To_Access_Conversions( Object=>st_mysql );
      use MYSQL_Access;

      ptr : Object_Pointer := null;
      ---
   begin
      ptr := To_Pointer( server );
      return Integer(ptr.Affected_Rows);
   end Num_Affected;

   ---======================================================================---
   ---===             C O M P O  N E N T    I N T E R F A C E            ===---
   ---======================================================================---
   ---======================================================================---
   ---===             C O M P O  N E N T    I N T E R F A C E            ===---
   ---======================================================================---

   --|
   --| Intialize the Object data and call the Initalization
   --| procedure of the extention.
   --|
   procedure Initialize( this  : in out Object ) is
   begin
      if this.data /= null then
         raise Usage_Error;
      end if;

      this.data := new Object_Data;

   end Initialize;

   --|
   --| Finalize the instance by releasing the object resources
   --|
   procedure Finalize( this : in out Object ) is
      data   : Object_Data_Access renames this.data;

      procedure Free is
            new Unchecked_Deallocation( Object_Data, Object_Data_Access);
   begin
      if data = null then
         raise Not_Initialized;
      end if;

      if data.connected = true  then
         Disconnect( this );
      end if;

      for i in Query_ID loop
         if data.query(i).result /= Null_Address then
	    Drop_Query( this, i );
	 end if;
      end loop;

      Free( data );

      this.data := null;

   end Finalize;

   ---=====================================================================---
   ---===           A T T R I B U T E    F U N C T I O N S              ===---
   ---=====================================================================---

   --|
   --| Set the user Name for a server connection
   --|
   procedure User(
      this   : in out Object;
      Name   : in String ) is
      ---
      data   : Object_Data_Access := Instance( this, "User");
      ---
   begin
      data.Login_Name := To_Unbounded_String( Name );
   end User;

   procedure Password(
      this   : in out Object;
      pwd    : in String ) is
      ---
      data : Object_Data_Access := Instance( this, "Password" );
      ---
   begin
       data.Password := To_Unbounded_String( pwd );
   end Password;

   ---=====================================================================---
   ---===                        M E T H O D S                          ===---
   ---=====================================================================---

   --|
   --| Connect to the server specified by the Object.
   --|
   procedure Connect(
      this   : in out Object;
      Server : in String ) is
      ---
      data   : Object_Data_Access := Instance( this, "Connect" );
      ---
   begin
      Data.Server_Name := To_Unbounded_String( Server );
      Data.Server := MYSQL_Connect( Null_Address,
                 New_String( To_String( data.Server_Name) ),
                 New_String( To_String(data.Login_Name) ),
                 New_String( To_String(data.Password) ) );

      if Data.Server = Null_Address then
         raise Connect_Failure;
      end if;

      Data.Connected := True;
   end Connect;

   --|
   --| Disconnect from a server.
   --|
   procedure Disconnect(
      this   : in out Object ) is
      ---
      data   : Object_Data_Access := Instance( this, "Disconnect");
      ---
   begin
      mysql_close( data.Server );
      data.Connected := False;

   end Disconnect;

   --|
   --| Select a data base from the server.
   --|
   procedure Select_DB(
      this   : in out Object;
      name   : in String ) is
      ---
      data   : Object_Data_Access := Instance( this, "Select_DB" );
      RC     : C.Int := 0;
      ---
   begin

      RC :=  mysql_select_db( data.Server, New_String( name ));
      if not ( RC = 0 ) then
         raise DB_Name_Failure;
      end if;

   end Select_DB;

   --|
   --|  Return the number of rows found during a selection
   --|
   function Nbr_Of_Rows(
      this       : in Object;
      id         : in Query_ID ) return Integer is
      ---
      Data : Object_Data_Access := Instance( this, "Nbr_Of_Rows");
      ---
   begin
      return NumRows(Data.Query(id).Result );
   end Nbr_Of_Rows;

   --|
   --| Perform a query on the server and retrieve the result
   --| context. The function returns the id of the query context.
   --|
   function Query(
      this       : in Object;
      query      : in String ) return Query_ID is
      ---
      Data       : Object_Data_Access := Instance( this, "Query");
      Result     : Query_ID := 1;
      RC         : C.int    := 0;
      Slot_Found : Boolean  := False;
      RES_Address: MYSQL_RES;
      ---
   begin

      -- perform the query
      RC := mysql_query( data.Server, New_String( query ));
      if RC /= 0  then
         raise No_Query_Result;
      end if;

      -- get a result identifier and the field information
      for i in Query_ID loop
         if data.Query(i).In_Use = False then
            Result := i;
            Data.Query(i).In_Use := True;

            RES_Address := mysql_store_result( data.Server );
            if RES_Address /= Null_Address then

               if NumRows( RES_Address ) < 1 then
                  raise No_Query_Result;
               end if;

               data.Query(i).Result := RES_Address;
               data.Query(i).Row    := mysql_fetch_row(data.Query(i).Result);
               for i in Field_Number loop
                  declare
                     Field : MYSQL_Field := mysql_fetch_field( Data.Query(Result).Result  );
                  begin
                     exit when Field = null;

                     data.Query( Result ).Field(i) := Field.all;
                  end;
               end loop;
            else
               data.Query(i).Affected_Rows := Num_Affected( Data.Server);
            end if;

            Slot_Found := True;
            exit;

         end if;
      end loop;

      if Slot_Found then
         return Result;
      else
         raise Out_Of_Resources;
      end if;

   end Query;

   --|
   --| Release the query context
   --|
   procedure Drop_Query(
      this    : in out Object;
      id      : in Query_ID ) is
      ---
      Data    : Object_Data_Access := Instance( this, "Drop_Query" );
      Q       : Query_Context renames Data.Query(id);
      ---
   begin
      if Q.In_Use then
         if Q.Result /= Null_Address then
            mysql_free_result( Q.Result);
         end if;
         Q.Result := Null_Address;
         Q.In_Use := False;
      end if;
      ---
   end Drop_Query;

   --|
   --| Get the next matching tuple for the given query
   --|
   procedure Next(
      this    : in out Object;
      query   : in Query_ID ) is
      ---
      Data    : Object_Data_Access := Instance( this, ".Next");
      Q       : Query_Context    renames Data.Query(query);
      ---
   begin
      Validate( data, query );

      if Q.Result /= Null_Address then
         Q.Row := mysql_fetch_row( Q.Result );
         if Q.Row = Null_Address then
            raise No_More_Rows;
         end if;
      else
         raise Usage_Error;
      end if;
      ---
   end Next;

   --|
   --| Translate the name into a field number.
   --|
   function Get_Field_Number(
      this       : Object;
      id         : Query_ID; name : String) return Field_Number is
      ---
      data       : Object_Data_Access := Instance( this, "Get_Field_Number" );
      ---
   begin
      Validate( data, id );

      for i in Field_Number loop
         if Value( Data.Query(id).Field(i).Name ) = name then
            return i;
         end if;
      end loop;

      raise Unknown_Field;
   end Get_Field_Number;

   --|
   --| Get String field identified by the field number
   --|
   function String_Field(
      this       : in Object;
      id         : in Query_ID;
      fNumber    : in Field_Number ) return Unbounded_String is
      ---
      data       : Object_Data_Access := Instance( this, "String_Field");
      length     : Positive;
      Q          : Query_Context renames  data.Query(id);
      ---
   begin
      Validate( data, id );

      length := Positive( Q.Field(fNumber).length );

      if Q.Field(fNumber).c_type = FIELD_TYPE_STRING or
         Q.Field(fNumber).c_type = FIELD_TYPE_VAR_STRING then
         return
            To_Unbounded_String(
               Copy_String( Get_Address( Q.Row, fNumber ), length ));
      else
         raise Field_Type_Error;
      end if ;
   end String_Field;

   --|
   --| Get the named string field
   --|
   function String_Field(
      this      : in Object;
      id        : in Query_ID;
      name      : in String ) return Unbounded_String is
      ---
      fNumber : Field_Number;
      ---
   begin
      fNumber := Get_Field_Number( this, id, name );
      return String_Field( this, id, fNumber );
   end String_Field;

   --|
   --| Get String field identifird by the field number
   --|
   function Date_Field(
      this    : in Object;
      id      : in Query_ID;
      fNumber : in Field_Number ) return Time is
      ---
      data    : Object_Data_Access := Instance( this, "Date_Field");
      length  : Positive;
      Q       : Query_Context renames Data.Query(id);
      ---
   begin
      Validate( data, id );

      length  := Positive( Q.Field(fNumber).length );

      if Q.Field(fNumber).c_type = FIELD_TYPE_DATE then
         declare
            Time_String : String(1..length);
            Year        : Year_Number;
            Month       : Month_Number;
            Day         : Day_Number;
            tmp         : String(1..5);

            next        : Positive := 1;

            -- copy a date element into the tmp string and return the start offset
            procedure Copy_Element( Str : String ) is
            begin
               tmp := (others => ' ');

               for i in 1..tmp'Length loop
                  if Str(next) = '-' then
                     next := next + 1;
                     exit;
                  end if;
                  tmp(i) := Str(next);

                  exit when next = length;
                  next := next + 1;
               end loop;

            exception
               when Others =>
                  put_line("Error in Copy_Element : " & Str & " at " & Integer'Image(next) );
                  raise;

            end Copy_Element;

         begin
            Move( Source => Copy_String( Get_Address( data.Query(id).Row, fNumber ), length ),
                  Target => Time_String);

            Copy_Element( Time_String );
            Year  := Year_Number'Value( tmp );


            Copy_Element( Time_String );
            Month := Month_Number'Value( tmp );

            Copy_Element( Time_String );
            Day   := Day_Number'Value( tmp );

            return Time_Of( Year, Month, Day );

         exception
            when The_Error : Others =>
                raise Field_Parse_Error;
         end;
      else
         raise Field_Type_Error;
      end if;

   end Date_Field;

   --|
   --| Get the named string field
   --|
   function Date_Field(
      this    : in Object;
      id      : in Query_ID;
      name    : in String ) return Time is
      ---
      fNumber : Field_Number;
      ---
   begin
      fNumber := Get_Field_Number( this, id, name );

      return Date_Field( this, id, fNumber );
   end Date_Field;

   --|
   --| return the blob field given by the columne number
   --|
   function Blob_Field (
      this    : in Object;
      id      : in Query_ID; fNumber : in  Field_Number ) return Blob is
      ---
      data    : Object_Data_Access := Instance( this, "Blob_Field");
      length  : Positive;
      Q       : Query_Context renames Data.Query(id);
      ---
   begin
      Validate( data, id );

      length  := Positive( Q.Field(fNumber).length );

      if Q.Field(fNumber).c_type = FIELD_TYPE_BLOB or
         Q.Field(fNumber).c_type = FIELD_TYPE_TINY_BLOB or
         Q.Field(fNumber).c_type = FIELD_TYPE_MEDIUM_BLOB or
         Q.Field(fNumber).c_type = FIELD_TYPE_LONG_BLOB or
         Q.Field(fNumber).c_type = FIELD_TYPE_BLOB
      then
         return Copy_String(
            Get_Address( data.Query(id).Row, fNumber ),
            length );
      else
         raise Field_Type_Error;
      end if;

   end Blob_Field;

   --|
   --| return the named blob field
   --|
   function Blob_Field(
      this    : in Object;
      id      : in Query_ID;
      name    : in String ) return Blob is
      ---
      fNumber : Field_Number;
      ---
   begin
      fNumber := Get_Field_Number( this, id, name );

      return Blob_Field( this, id, fNumber );
   end Blob_Field;

   --|
   --|  return a number field as an integer.
   --|
   function Integer_Field(
      this    : in Object;
      id      : in Query_ID;
      fNumber : in Field_Number ) return Integer is
      ---
      data    : Object_Data_Access := Instance( this, "Integer_Field");
      length  : Positive;
      Q       : Query_Context renames Data.Query(id);
      ---
   begin
      Validate( data, id );

      length  := Positive( Data.Query(id).Field(fNumber).length );

      if Q.Field(fNumber).c_type = FIELD_TYPE_TINY or                                    -- mysql_com.h:64
         Q.Field(fNumber).c_type = FIELD_TYPE_SHORT or
         Q.Field(fNumber).c_type = FIELD_TYPE_LONG then
         return Integer'Value(
            Copy_String( Get_Address(data.Query(id).Row, fNumber),length));
      else
         raise Field_Type_Error;
      end if;
   end Integer_Field;

   --|
   --| return the number field by name
   --|
   function Integer_Field(
      this    : in Object;
      id      : in Query_ID; name : in String ) return Integer is
      ---
      fNumber : Field_Number;
      ---
   begin
      fNumber := Get_Field_Number( this, id, name );

      return Integer_Field( this, id, fNumber );

   end Integer_Field;

   --|
   --| return a float field
   --|
   function Float_Field(
      this    : in Object;
      id      : in Query_ID;
      fNumber : in Field_Number ) return Float is
      ---
      data    : Object_Data_Access := Instance( this, ".Float_Field");
      length  : Positive;
      Q       : Query_Context    renames Data.Query(id);
      ---
   begin
      Validate( data, id );

      length  := Positive( Q.Field(fNumber).length );

      if Q.Field(fNumber).c_type = FIELD_TYPE_FLOAT or                                    -- mysql_com.h:64
         Q.Field(fNumber).c_type = FIELD_TYPE_DOUBLE then
         return Float'Value(
            Copy_String( Get_Address(data.Query(id).Row, fNumber),length));
      else
         raise Field_Type_Error;
      end if;
   end Float_Field;

   --|
   --| return the number field by name
   --|
   function Float_Field(
      this    : in Object;
      id      : in Query_ID; name : in String ) return Float is
      ---
      fNumber : Field_Number;
      ---
   begin
      fNumber := Get_Field_Number( this, id, name );
      return Float_Field( this, id, fNumber );
   end Float_Field;

end DB;

